#!/usr/bin/env python3
# coding: utf-8

import sqlite3

from multiprocessing import Pool
from napoleontoolbox.file_saver import dropbox_file_saver

class SignalAggregatingLauncher():
    def __init__(self,drop_token='', dropbox_backup=True, local_root_directory='../data',user='napoleon',  db_path_suffix = '_run.sqlite'):
        self.local_root_directory = local_root_directory
        self.user = user
        self.db_path_suffix = db_path_suffix
        self.filename =  user + db_path_suffix
        self.db_path = self.local_root_directory + self.filename
        self.runs = []
        self.totalRow = 0
        self.instantiateTable()
        self.empty_runs_to_investigate = []
        self.dbx = dropbox_file_saver.NaPoleonDropboxConnector(drop_token=drop_token,dropbox_backup=dropbox_backup)


    def launchParallelPool(self,runner,use_num_cpu):
        print('launching parallel computation for alphas at ')
        print('launching parallel computation')
        with Pool(processes=use_num_cpu) as pool:
            run_results = pool.map(runner.runSingleTrial, runner.signals_list)
        print('parallel computation done')
        print('results length')
        print(len(run_results))

        aggregated_df = None
        counter = 0
        mapping_dictionary ={}
        for me_result in run_results:
            if me_result is not None:
                signal_column = 'signal' + str(counter)
                if aggregated_df is None:
                    aggregated_df = me_result.copy()
                    aggregated_df = aggregated_df.rename(columns={runner.signals_list[counter]: signal_column})
                else:
                    aggregated_df[signal_column] = me_result[runner.signals_list[counter]].copy()
                mapping_dictionary[runner.signals_list[counter]] = [signal_column]
            counter = counter + 1


        runner.saving_cumulated_signals(aggregated_df, mapping_dictionary)

    def launchSequential(self, toRun):
        toRun(self)

    def instantiateTable(self):
        sqliteConnection = None
        try:
            sqliteConnection = sqlite3.connect(self.db_path)
            sqlite_create_table_query = '''CREATE TABLE parallel_run (
                                        effective_date date PRIMARY KEY);'''
            cursor = sqliteConnection.cursor()
            print("Successfully Connected to SQLite")
            cursor.execute(sqlite_create_table_query)
            sqliteConnection.commit()
            print("SQLite table created")
            cursor.close()
        except sqlite3.Error as error:
            print("Error while creating a sqlite table", error)
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("sqlite connection is closed")

    def addRunTurnOver(self,run):
        sqliteConnection = None
        try:
            sqliteConnection = sqlite3.connect(self.db_path)
            sqlite_create_table_query = 'alter table parallel_run add column ' + run + '_turn_over real'
            cursor = sqliteConnection.cursor()
            print("Successfully Connected to SQLite")
            cursor.execute(sqlite_create_table_query)
            sqliteConnection.commit()
            print("SQLite table created")
            cursor.close()
            self.runs.append(run)
        except sqlite3.Error as error:
            print("Error while creating a sqlite table", error)
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("sqlite connection is closed")  # adding the column

    def addRunPerf(self,run):
        sqliteConnection = None
        try:
            sqliteConnection = sqlite3.connect(self.db_path)
            sqlite_create_table_query = 'alter table parallel_run add column ' + run + '_perf_return real'
            cursor = sqliteConnection.cursor()
            print("Successfully Connected to SQLite")
            cursor.execute(sqlite_create_table_query)
            sqliteConnection.commit()
            print("SQLite table created")
            cursor.close()
            self.runs.append(run)
        except sqlite3.Error as error:
            print("Error while creating a sqlite table", error)
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("sqlite connection is closed")  # adding the column

    def insertResults(self, run, values):
        sqliteConnection = None
        try:
            success = True
            sqliteConnection = sqlite3.connect(self.db_path)
            cursor = sqliteConnection.cursor()
            for i, v in values.iteritems():
                sqlite_insert_query = """INSERT INTO 'parallel_run' ('effective_date', '""" + run + """') VALUES ('""" + str(
                    i) + """','""" + str(v) + """');"""
                cursor.execute(sqlite_insert_query)
            sqliteConnection.commit()
            cursor.close()
        except sqlite3.Error as error:
            print("Error while working with SQLite", error)
            success = False
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("The Sqlite connection is closed")
        return success

    def updateResults(self, run, values):
        sqliteConnection = None
        try:
            sqliteConnection = sqlite3.connect(self.db_path)
            cursor = sqliteConnection.cursor()
            for i, v in values.iteritems():
                sqlite_update_query = """UPDATE 'parallel_run' set '""" + run + """' = '""" + str(
                    v) + """' where effective_date = '""" + str(i) + """'"""
                cursor.execute(sqlite_update_query)
            sqliteConnection.commit()
            cursor.close()
        except sqlite3.Error as error:
            print("Error while working with SQLite", error)
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("The Sqlite connection is closed")

    def insertPerfTurnOverResults(self, run, signal_results_df):
        sqliteConnection = None
        try:
            success = True
            sqliteConnection = sqlite3.connect(self.db_path)
            cursor = sqliteConnection.cursor()
            for date_index, row in signal_results_df.iterrows():
                sqlite_insert_query = """INSERT INTO 'parallel_run' ('effective_date', '""" +run +"""_perf_return','"""+run +"""_turn_over')"""+""" VALUES ('""" + str(
                    date_index) + """','""" + str(row['perf_return']) + """','""" + str(row['turn_over']) + """');"""
                cursor.execute(sqlite_insert_query)
            sqliteConnection.commit()
            cursor.close()
        except sqlite3.Error as error:
            print("Error while working with SQLite", error)
            success = False
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("The Sqlite connection is closed")
        return success

    def updatePerfTurnOverResults(self, run, signal_results_df):
        sqliteConnection = None
        try:
            sqliteConnection = sqlite3.connect(self.db_path)
            cursor = sqliteConnection.cursor()
            for date_index, row in signal_results_df.iterrows():
                sqlite_update_query = """UPDATE 'parallel_run' set ('""" + run + """_perf_return','""" + run + """_turn_over')""" + """ = ('""" + str(
                    row['perf_return']) + """','""" + str(row['turn_over']) + """') where effective_date = '""" + str(
                    date_index) + """'"""
                cursor.execute(sqlite_update_query)
            sqliteConnection.commit()
            cursor.close()
        except sqlite3.Error as error:
            print("Error while working with SQLite", error)
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("The Sqlite connection is closed")

    def addPerfTurnOverResults(self, run, values):
        success = self.insertPerfTurnOverResults(run,values)
        if not success :
            self.updatePerfTurnOverResults(run, values)


    def addResults(self, run, values):
        success = self.insertResults(run,values)
        if not success :
            self.updateResults(run, values)

    def addRun(self,run):
        sqliteConnection = None
        try:
            sqliteConnection = sqlite3.connect(self.db_path)
            sqlite_create_table_query = 'alter table parallel_run add column ' + run + ' real'
            cursor = sqliteConnection.cursor()
            print("Successfully Connected to SQLite")
            cursor.execute(sqlite_create_table_query)
            sqliteConnection.commit()
            print("SQLite table created")
            cursor.close()
            self.runs.append(run)
        except sqlite3.Error as error:
            print("Error while creating a sqlite table", error)
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("sqlite connection is closed")  # adding the column

    def saveResults(self, run, strat_results_df):
        self.addRun(run)
        #self.addRunPerf(run)
        #self.addRunTurnOver(run)
        self.addResults(run, strat_results_df)


    def checkRunExistence(self,run):
        sqliteConnection = None
        runExistence=None
        try:
            sqliteConnection = sqlite3.connect(self.db_path)
            sqlite_select_run_query = 'SELECT effective_date,'+run + ' FROM parallel_run'
            cursor = sqliteConnection.cursor()
            print("Successfully Connected to SQLite")
            cursor.execute(sqlite_select_run_query)
            sqliteConnection.commit()
            print("SQLite table created")
            cursor.close()
            runExistence = True
        except sqlite3.Error as error:
            print("The run is not present", error)
            runExistence = False
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("sqlite connection is closed")  # adding the column
        return runExistence

    def datesCount(self):
        sqliteConnection = None
        try:
            sqliteConnection = sqlite3.connect(self.db_path, timeout=20)
            cursor = sqliteConnection.cursor()
            sqlite_select_query = """SELECT count(*) from parallel_run"""
            cursor.execute(sqlite_select_query)
            totalRows = cursor.fetchone()
            print("Total rows are:  ", totalRows)
            cursor.close()
            return totalRows
        except sqlite3.Error as error:
            print("Failed to read data from sqlite table", error)
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("The Sqlite connection is closed")



    def upload_results_to_dropbox(self):
        print('uploading to dropbox the results '+self.db_path)
        self.dbx.uploadFileToDropbox(filename=self.filename,fullpath = self.db_path)



def get_database_name(user, frequence, underlying, algo, starting_date, running_date, db_path_suffix):
    return  user + '_' + frequence + '_' + underlying + '_' + algo + '_'+ starting_date.strftime(
        '%d_%b_%Y') + '_' + running_date.strftime('%d_%b_%Y') + db_path_suffix


import numpy as np
import scipy
import pandas as pd
from scipy.stats import linregress

from napoleontoolbox.signal import signal_utility

from sklearn import preprocessing
from sklearn.preprocessing import StandardScaler
from functools import partial
from numpy.lib.stride_tricks import as_strided as stride
import pandas as pd
from scipy.optimize import Bounds, LinearConstraint, minimize
from napoleontoolbox.utility import metrics
import numpy as np
from napoleontoolbox.signal import signal_utility
from napoleontoolbox.connector import napoleon_connector
import json

from napoleontoolbox.parallel_run import signal_result_analyzer
from sklearn.metrics import accuracy_score
from functools import partial


class SimpleThresholdModel():

    def __init__(self, nb_selected=10, number_per_year=365):
        self.model = None
        self.seed = 0
        self.nb_selected = nb_selected
        self.number_per_year = number_per_year

    def calibrate(self, X, y):
        return

    def fit(self, X_train, y_train, X_val, y_val):
        print(f'finding the {self.nb_selected} most correlated features with the output')
        output_correlations = []
        X = X_train.copy().fillna(0.)
        y = y_train.copy().fillna(0.)
        for daily_feat in X_train.columns:
            #print(daily_feat)
            #print(f'number of distinct {X[daily_feat].nunique()} values')
            correlation_matrix = np.corrcoef(X[daily_feat], y)
            output_correlations.append(
                {
                    'feature': daily_feat,
                    'correlation': correlation_matrix[0][1]
                }
            )
        output_correlations_df = pd.DataFrame(output_correlations)
        output_correlations_df['abs_correlation'] = abs(output_correlations_df['correlation'])
        output_correlations_df = output_correlations_df.sort_values(by='abs_correlation', ascending=False)
        output_correlations_df['effect_sign']=np.sign(output_correlations_df['correlation'])
        output_correlations_df.index = output_correlations_df.feature
        selected_output_correlations_df = output_correlations_df.head(self.nb_selected)
        self.selected_features = list(selected_output_correlations_df.feature)
        correlation_effects = selected_output_correlations_df['effect_sign'].to_dict()
        selected_X = X[self.selected_features].copy()
        scaler = StandardScaler()
        scaler.fit(selected_X)
        transformed_selected_X = scaler.transform(selected_X)
        transformed_selected_X_df = pd.DataFrame(data = transformed_selected_X, columns = selected_X.columns, index = selected_X.index)
        #print(transformed_selected_X.shape)
        threshold_dic = {}
        for feature_to_investigate in self.selected_features:
            #print(feature_to_investigate)
            signal_way = correlation_effects[feature_to_investigate]
            x = transformed_selected_X_df[feature_to_investigate].values.copy()
            np.random.seed(self.seed)
            def find_best_threshold(number_per_year, way,y,x,w):
                if np.sign(way)>0:
                    y_pred_discrete = x > w[0]
                else:
                    y_pred_discrete = x < w[0]
                perf_df = signal_utility.reconstitute_prediction_perf(y_pred=y_pred_discrete, y_true=y,
                                                                      transaction_cost=False,
                                                                      print_turnover=False)
                sharpe_strat = metrics.sharpe(perf_df['perf_return'].dropna(), period=number_per_year, from_ret=True)
                return -1*sharpe_strat
            f_to_optimize = partial(find_best_threshold, self.number_per_year, signal_way, y, x)
            N_ = 1
            w0 = np.array([0.5])
            # Set constraints
            # const_sum = LinearConstraint(np.ones([1, N_]), [1], [1])
            up_bound_ = 1.
            low_bound_ = 0.
            const_ind = Bounds(low_bound_ * np.ones([N_]), up_bound_ * np.ones([N_]))
            splitting_quantile_threshold = 0.5
            trouble = False
            try:
                # Optimize f
                conv_results = minimize(
                    f_to_optimize,
                    w0,
                    method='SLSQP',
                    # constraints=[const_sum],
                    bounds=const_ind
                )
                w__=conv_results.x
                print(f'sharpe conv reults {conv_results.fun}')
                splitting_quantile_threshold = w__[0]
            except Exception as e:
                trouble = True
                ## we should print
                print(e)
            if np.isnan(splitting_quantile_threshold):
                splitting_quantile_threshold = 0.5
            threshold_dic[feature_to_investigate] = splitting_quantile_threshold
            #print('done')
        self.threshold_dic = threshold_dic
        self.correlation_effects = correlation_effects
        self.scaler = scaler
        return

    def predict(self, X_test):
        X = X_test.copy()
        selected_X = X[self.selected_features].copy()
        transformed_selected_X = self.scaler.transform(selected_X)
        transformed_selected_X_df = pd.DataFrame(data = transformed_selected_X, columns = selected_X.columns, index = selected_X.index)
        for feature_to_investigate in self.selected_features:
            #print(feature_to_investigate)
            signal_way = self.correlation_effects[feature_to_investigate]
            x = transformed_selected_X_df[feature_to_investigate].values.copy()
            if np.sign(signal_way) > 0:
                y_pred_discrete = x > self.threshold_dic[feature_to_investigate]
            else:
                y_pred_discrete = x < self.threshold_dic[feature_to_investigate]
            transformed_selected_X_df[feature_to_investigate] = y_pred_discrete
        return transformed_selected_X_df.mean(axis=1)

    def get_features_importance(self, features_names):
        return



import numpy as np
import pandas as pd

def most_correlated_features(X,threshold):
    generic_features = ['close','high', 'low', 'open', 'volume']
    columns_to_compute = [col for col in X.columns if col not in generic_features]
    output_correlations = []
    for i in range(len(columns_to_compute)):
        for j in range(i+1,len(columns_to_compute)):
            left_feature = columns_to_compute[i]
            right_feature=columns_to_compute[j]
            if left_feature != right_feature:
                # print(left_feature, right_feature)
                # print(X[left_feature].nunique())
                # print('done')
                correlation_matrix = np.corrcoef(X[left_feature],X[right_feature])
                output_correlations.append(
                    {
                        'daily_feature_1':left_feature,
                        'daily_feature_2':right_feature,
                        'correlation':correlation_matrix[0][1]
                    }
                )

    output_correlations_df = pd.DataFrame(output_correlations)
    output_correlations_df['abs_correlation'] = abs(output_correlations_df['correlation'] )
    output_correlations_df = output_correlations_df.sort_values(by='abs_correlation', ascending=False)
    columns_to_drop = output_correlations_df.loc[output_correlations_df['abs_correlation'] > threshold]['daily_feature_2'].unique()
    return columns_to_drop


def get_binarization_threshold(X):
        return X.mean().to_dict()

def binarize_features(X,features_threshold):
    X[list(features_threshold.keys())] = (X[features_threshold.keys()] >= list(features_threshold.values())).astype(int)
    return X


from napoleontoolbox.forecasting import preprocessing_features
import numpy as np
import pandas as pd

class BaselineModel():

    def __init__(self, threshold, number_per_year=252):
        self.model = 'simple_model'
        self.threshold = threshold
        self.output_ways = []
        self.last_avg_return = pd.DataFrame()

    def fit(self, X_train, y_train, X_val, y_val):#, method = 'standard'):
        self.columns_to_drop = preprocessing_features.most_correlated_features(X_train, self.threshold)
        X_train = X_train.drop(self.columns_to_drop, axis=1)
        X_val = X_val.drop(self.columns_to_drop, axis=1)
        way_last_return = np.sign(y_train[-1])

        for daily_feat in X_train.columns:
            # print(daily_feat)
            # print(f'number of distinct {X[daily_feat].nunique()} values')
            self.last_avg_return = pd.DataFrame(pd.DataFrame(y_train).mean())
            correlation_matrix = np.corrcoef(X_train[daily_feat], y_train)
            if (way_last_return>0 and correlation_matrix[0][1]>0):
                self.output_ways.append(1)

            elif (way_last_return < 0 and correlation_matrix[0][1] < 0):
                self.output_ways.append(0)
            else:
                self.output_ways.append(0.5)

    def predict(self, X_test):
        X_test = X_test.drop(self.columns_to_drop, axis=1)
        self.last_avg_return = self.last_avg_return.append([self.last_avg_return] * (len(X_test) - 1), ignore_index=True)
        next_way = np.max(self.output_ways)
        if next_way>0:
            y_pred = self.last_avg_return
        elif next_way<0:
            y_pred = 0*self.last_avg_return
        else:
            y_pred=0.5*self.last_avg_return
        return y_pred.values

    def get_features_importance(self, features_names):
        run_importances = {}
        # for (name, imp) in zip(features_names, self.model.coef_):
        #     run_importances[name] = imp
        return run_importances


#!/usr/bin/env python3
# coding: utf-8

""" Function to manage date and time. """

# Built-in packages
import datetime

# Third party packages

# Local packages


__all__ = ['parser_date']

# Set handler
HANDLER_DAY = {'y': 365, 'q': 91, 'm': 30, 'w': 7, 'd': 1}


def _parse_date(i, idx):
    # Parser of date
    if isinstance(i, int):
        t = idx[max(min(i, idx.size), 0)]

    elif isinstance(i, datetime.date):
        t = i

    elif isinstance(i, str):
        t = datetime.date(i.split('-'))

    else:
        raise ValueError("{} type not allowed, only int, datetime.date or str\
                          as '%Y-%m-%d' format.".format(type(i)))

    return t


def parser_date(date, period='ytd'):
    """ Compute the lower bound of `period` until `date`.

    Parameters
    ----------
    date : datetime.datetime
        Date of the upper bound of the `period`.
    period : {'ytd', 'qtd', 'mtd', 'y', 'q', 'm', 'w', 'd'}
        - 'ytd' is year to date.
        - 'qtd' is quarter to date.
        - 'mtd' is month to date.
        - 'y', 'q', 'm', 'w', 'd' are respectively one unit of year, quarter,
        month, week, and day. You can also add a number in front of the unit.

    Returns
    -------
    datetime.datetime
        Date of the lower bound of the `period`.

    """
    y, m, d = (getattr(date, p) for p in ['year', 'month', 'day'])
    if period.lower()[-2:] == 'td':
        n, u = _handler_unit_period(period[:-2])
        date = date - datetime.timedelta(HANDLER_DAY[u] * (n - 1))

        return _parser_todate(date, u)

    else:
        n, u = _handler_unit_period(period)

        return date - datetime.timedelta(HANDLER_DAY[u] * n)


def _parser_todate(date, u):
    if u == 'm':
        m = date.month

    elif u == 'q':
        m = date.month - (date.month % 3)  # - 1)

    elif u == 'y':
        m = 1

    else:
        raise ValueError("Only {ytd, qtd, mtd} are allowed.")

    if date.month == m and date.day <= 4:

        return _parser_todate(date - datetime.timedelta(days=4), u=u)

    return date.replace(month=m, day=1)


def _handler_unit_period(period):
    """ Split the period into number and unit.

    Parameters
    ----------
    period : str
        A number plus a period unit {'y', 'q', 'm', 'w', d'}.

    Returns
    -------
    int
        Number of unit period.
    str
        Symbol of unit period.

    """
    if len(period) == 0:

        raise ValueError(
            "Period must be a string such that number + {'y','q','m','w','d'}"
        )

    elif len(period) == 1 and period.lower() in ['y', 'm', 'd', 'q', 'w']:

        return 1, period

    elif period.lower()[-1] in ['y', 'm', 'd', 'q', 'w']:

        return int(period[:-1]), period[-1]

    else:

        raise ValueError(
            "Period must be a string such that number + {'y','q','m','w','d'}"
        )



#!/usr/bin/env python3
# coding: utf-8

import numpy as np
from sklearn.metrics import mean_squared_error
from napoleontoolbox.roller import roller
import pickle
import shap
import pandas as pd
import torch.nn
import torch
import torch.nn as nn
from torch.autograd import Variable

from napoleontoolbox.callback import ts_callback
from functools import partial

class _ShortCutLSTMBaseNeuralNet(torch.nn.Module):
    """ Base object for neural network model with PyTorch.
    Inherits of torch.nn.Module object with some higher level methods.
    Attributes
    ----------
    criterion : torch.nn.modules.loss
        A loss function.
    optimizer : torch.optim
        An optimizer algorithm.
    N, M : int
        Respectively input and output dimension.
    Methods
    -------
    set_optimizer
    train_on
    predict
    set_data
    See Also
    --------
    MultiLayerPerceptron, RollingBasis
    """

    def __init__(self):
        """ Initialize. """
        torch.nn.Module.__init__(self)

    def set_optimizer(self, criterion, optimizer, lr_type, **kwargs):
        """ Set the optimizer object.
        Set optimizer object with specified `criterion` as loss function and
        any `kwargs` as optional parameters.
        Parameters
        ----------
        criterion : torch.nn.modules.loss
            A loss function.
        optimizer : torch.optim
            An optimizer algorithm.
        **kwargs
            Keyword arguments of `optimizer`, cf PyTorch documentation [1]_.
        Returns
        -------
        BaseNeuralNet
            Self object model.
        References
        ----------
        .. [1] https://pytorch.org/docs/stable/optim.html
        """
        self.criterion = criterion()
        self.optimizer = optimizer(self.parameters(), **kwargs)
        self.lr = None
        lr = kwargs['lr']

        sched_standard = ts_callback.create_standard_annealing_profile(lr)
        cbf_func = partial(ts_callback.ParamScheduler, 'lr', sched_standard,lr_type)

        cb = cbf_func()
        setattr(self, cb.name, cb)
        cb.set_runner(self)
        self.cbs.append(cb)

        return self

    @torch.enable_grad()
    def train_on(self, X, X_shortcut, y):
        """ Trains the neural network model.
        Parameters
        ----------
        X, y : torch.Tensor
            Respectively inputs and outputs to train model.
        Returns
        -------
        torch.nn.modules.loss
            Loss outputs.
        """
        self.optimizer.zero_grad()
        outputs = self(X,X_shortcut)
        loss = self.criterion(outputs, y)
        # if self.num_layers>1:
        #     duplicate_y = torch.cat((y, y), 0)
        # else:
        #     duplicate_y = y
        # loss = self.criterion(outputs, duplicate_y)
        loss.backward()
        self.optimizer.step()
        self.loss = loss.item()
        return loss.item()

    @torch.no_grad()
    def predict(self, X, X_shortcut):
        """ Predicts outputs of neural network model.
        Parameters
        ----------
        X : torch.Tensor
           Inputs to compute prediction.
        Returns
        -------
        torch.Tensor
           Outputs prediction.
        """
        return self(X,X_shortcut).detach()

    def set_data(self, X, X_shortcut, y, x_type=None, y_type=None):
        """ Set data inputs and outputs.
        Parameters
        ----------
        X, y : array-like
            Respectively input and output data.
        x_type, y_type : torch.dtype
            Respectively input and ouput data types. Default is `None`.
        """
        assert x_type == y_type

        if x_type == torch.float64:
            self=self.double()

        if hasattr(self, 'N') and self.N != X.size(1):
            raise ValueError('X must have {} input columns'.foramt(self.N))

        if hasattr(self, 'M') and self.M != y.size(1):
            raise ValueError('y must have {} output columns'.format(self.M))

        self.X = self._set_data(X)
        self.X_shortcut = self._set_data(X_shortcut)
        self.y = self._set_data(y)

        if self.X.ndim == 2:
            self.T, self.N = self.X.size()
            T_veri, self.M = self.y.size()
        if self.X.ndim == 3:
            self.T,self.H, self.N = self.X.size()
            T_veri, self.M = self.y.size()

        if self.T != T_veri:
            raise ValueError('{} time periods in X differents of {} time \
                             periods in y'.format(self.T, T_veri))

        return self

    def _set_data(self, X):
        """ Convert array-like data to tensor. """
        # TODO : Verify dtype of data torch tensor
        if isinstance(X, np.ndarray):
            return torch.from_numpy(X)

        elif isinstance(X, pd.DataFrame):
            # TODO : Verify memory efficiancy
            return torch.from_numpy(X.values)

        elif isinstance(X, torch.Tensor):

            return X

        else:
            raise ValueError('Unkwnown data type: {}'.format(type(X)))


class MultiLayerShortCutLSTM(_ShortCutLSTMBaseNeuralNet):
    r""" Neural network with MultiLayer Perceptron architecture.
    Refered as vanilla neural network model, with `n` hidden layers s.t
    n :math:`\geq` 1, with each one a specified number of neurons.
    Parameters
    ----------
    X, y : array-like
        Respectively inputs and outputs data.
    layers : list of int
        List of number of neurons in each hidden layer.
    activation : torch.nn.Module
        Activation function of layers.
    drop : float, optional
        Probability of an element to be zeroed.
    Attributes
    ----------
    criterion : torch.nn.modules.loss
        A loss function.
    optimizer : torch.optim
        An optimizer algorithm.
    n : int
        Number of hidden layers.
    layers : list of int
        List with the number of neurons for each hidden layer.
    f : torch.nn.Module
        Activation function.
    Methods
    -------
    set_optimizer
    train_on
    predict
    set_data
    See Also
    --------
    BaseNeuralNet, RollMultiLayerPerceptron
    """

    def __init__(self, X, X_shortcut, y, num_classes, input_size, hidden_size, num_layers,x_type=None, y_type=None, activation = None):
        """ Initialize object. """
        _ShortCutLSTMBaseNeuralNet.__init__(self)
        super(MultiLayerShortCutLSTM, self).__init__()

        self.set_data(X=X, X_shortcut=X_shortcut ,y=y, x_type=x_type, y_type=y_type)

        self.num_classes = num_classes
        self.num_layers = num_layers
        self.input_size = input_size
        self.shortcut_size = X_shortcut.shape[1]
        self.hidden_size = hidden_size

        self.lstm = nn.LSTM(input_size=input_size, hidden_size=hidden_size,
                            num_layers=num_layers, batch_first=True)

        self.fc = nn.Linear(hidden_size+self.shortcut_size, num_classes)
        # Set activation functions
        if activation is None:
            self.activation = lambda x: x
        else:
            self.activation = activation



    def forward(self, x, x_shortcut):
        # # reset the state of LSTM
        # # the state is kept till the end of the past sequence for the batch
        # h_batch = Variable(torch.zeros(
        #     self.num_layers, x.size(0), self.hidden_size))
        # c_batch = Variable(torch.zeros(
        #     self.num_layers, x.size(0), self.hidden_size))
        # # Propagate input through LSTM
        # ula, (h_out, _) = self.lstm(x, (h_batch, c_batch))

        ula, (h_out, _) = self.lstm(x)
        #clip = 0.001
        #torch.nn.utils.clip_grad_norm(self.lstm.parameters(), clip)

        h_out = h_out.view(-1, self.hidden_size)
        #print(h_out.shape)

        out = torch.cat([h_out, x_shortcut], dim=1)
        out = self.fc(out)

        #print(out.shape)
        out = self.activation(out)
        if self.num_layers>1:
            out = out[-x.shape[0]:]
        return out

def _type_convert(dtype):
    if dtype is np.float64 or dtype is np.float or dtype is np.double:
        return torch.float64

    elif dtype is np.float32:
        return torch.float32

    elif dtype is np.float16:
        return torch.float16

    elif dtype is np.uint8:
        return torch.uint8

    elif dtype is np.int8:
        return torch.int8

    elif dtype is np.int16 or dtype is np.short:
        return torch.int16

    elif dtype is np.int32:
        return torch.int32

    elif dtype is np.int64 or dtype is np.int or dtype is np.long:
        return torch.int64

    else:
        raise ValueError('Unkwnown type: {}'.format(str(dtype)))

class RollMultiLayerShortCutLSTM(MultiLayerShortCutLSTM, roller._ShortCutRollingBasis):
    """ Rolling version of the vanilla neural network model.
    TODO:
    - fix train and predict methods
    - finish docstring
    - finish methods
    """

    def __init__(self, X, X_shortcut, y, num_classes, input_size, hidden_size, num_layers, x_type = None, y_type = None, activation=None):
        """ Initialize rolling multi-layer perceptron model. """
        roller._ShortCutRollingBasis.__init__(self, X, X_shortcut, y)
        MultiLayerShortCutLSTM.__init__(self, X, X_shortcut, y, num_classes=num_classes, input_size=input_size,
                                hidden_size=hidden_size, num_layers=num_layers,x_type = x_type, y_type = y_type, activation = activation)
        assert x_type == y_type
        if x_type == torch.float64:
            self.double()

    def set_roll_period(self, train_period, test_period, start=0, end=None, repass_steps=1):
        """ Callable method to set target features data, and model.
        Parameters
        ----------
        train_period, test_period : int
            Size of respectively training and testing sub-periods.
        start : int, optional
            Starting observation, default is first observation.
        end : int, optional
            Ending observation, default is last observation.
        roll_period : int, optional
            Size of the rolling period, default is the same size of the
            testing sub-period.
        eval_period : int, optional
            Size of the evaluating period, default is the same size of the
            testing sub-period if training sub-period is large enough.
        batch_size : int, optional
            Size of a training batch, default is 64.
        epochs : int, optional
            Number of epochs, default is 1.
        Returns
        -------
        _RollingBasis
            The rolling basis model.
        """
        return roller._ShortCutRollingBasis.set_roll_period(
            self, train_period=train_period, test_period=test_period,
            start=start, end=end, repass_steps=repass_steps
        )

    def _train(self, X,X_shortcut,y):
        return self.train_on(X=X, X_shortcut=X_shortcut, y=y)

    def sub_predict(self, X, X_shortcut):
        """ Predict. """
        return self.predict(X=X,X_shortcut=X_shortcut)

    def eval_predictor_importance(self, features, features_names):
        explainer_shap = shap.GradientExplainer(model=self,
                                            data=features)
        # Fit the explainer on a subset of the data (you can try all but then gets slower)
        shap_values = explainer_shap.shap_values(X=features,
                                                 ranked_outputs=True)

        predictors_shap_values = shap_values[0]
        predictors_feature_order = np.argsort(np.sum(np.mean(np.abs(predictors_shap_values), axis=0), axis=0))

        predictors_left_pos = np.zeros(len(predictors_feature_order))

        predictors_class_inds = np.argsort([-np.abs(predictors_shap_values[i]).mean() for i in range(len(predictors_shap_values))])
        for i, ind in enumerate(predictors_class_inds):
            predictors_global_shap_values = np.abs(predictors_shap_values[ind]).mean(0)
            predictors_left_pos += predictors_global_shap_values[predictors_feature_order]

        predictors_ds = {}
        predictors_ds['features'] = np.asarray(features_names)[predictors_feature_order]
        predictors_ds['values'] = predictors_left_pos
        predictors_features_df = pd.DataFrame.from_dict(predictors_ds)
        values = {}
        for index, row in predictors_features_df.iterrows():
            values[row['features']]=row['values']
        return values

    def unroll(self, display_slices=False):
        self.handle_callback('begin_fit')
        for eval_slice, test_slice in self:
            # Compute prediction on eval and test set
            self.y_eval[eval_slice] = self.sub_predict(self.X[eval_slice],self.X_shortcut[eval_slice])
            predictors = self.X[test_slice]
            test_prediction = self.sub_predict(self.X[test_slice],self.X_shortcut[test_slice])
            if abs(test_prediction.numpy().sum())<= 1e-6:
                print('null prediction, investigate')
            self.y_test[test_slice] = test_prediction

            ev = self.y_eval[eval_slice]
            ev_true = self.y[eval_slice]

            tt = self.y_test[test_slice]
            tt_true = self.y[test_slice]

            self.loss_eval += [mean_squared_error(ev, ev_true)]
            self.loss_test += [mean_squared_error(tt, tt_true)]

            # Print loss on current eval and test set
            pct = (self.t - self.n - self.s) / (self.T - self.n - self.T % self.s)
            txt = '{:5.2%} is done | '.format(pct)
            txt += 'Eval loss is {:5.2} | '.format(self.loss_eval[-1])
            txt += 'Test loss is {:5.2} | '.format(self.loss_test[-1])
            if np.random.rand()>=0.8:
                print(txt)
            if display_slices:
                print('eval slice '+str(eval_slice))
                print('test slice ' + str(test_slice))
                print(self.X[test_slice])
                print(tt.sum())
                print(tt_true.sum())

        self.last_eval_slice = eval_slice
        self.last_test_slice = test_slice
        return eval_slice, test_slice, predictors,  tt ,  tt_true

    def print_parameters(self):
        # Print model's state_dict
        print("to be done")

    def save_model_and_last_state(self, state=None, model_path=''):
        file_obj = open(model_path+'_py.obj', 'wb')
        pickle.dump(state, file_obj)
        file_obj.close()

    def resume_state(self, model_path=''):
        file_obj = open(model_path+'_py.obj', 'rb')
        self.state = pickle.load(file_obj)
        return self.state


#!/usr/bin/env python3
# coding: utf-8

import requests
import pandas as pd
import math
import datetime
from napoleontoolbox.connector import napoleon_connector


def convert_string(s):
    from dateutil import parser
    dt_object = parser.parse(s)
    return dt_object.timestamp()

def convert_datetime(d):
    return str(d.date())+'T00:00:00Z'


def get_events_augmento(currency='bitcoin', startdate='2018-01-01T00:00:00Z', enddate='2019-10-01T00:00:00Z', source='twitter', precision='24H', local_root_directory=None):
    r = requests.request("GET", "http://api-dev.augmento.ai/v0.1/coins")
    # print('Coins available\n', r.content, '\n')
    r = requests.request("GET", "http://api-dev.augmento.ai/v0.1/sources")
    # print('Sources available\n', r.content, '\n')
    topics = requests.request("GET", "http://api-dev.augmento.ai/v0.1/topics").json()
    # print('Topics available\n', topics, '\n')
    topics = topics.values()

    startdate_unix, enddate_unix = startdate.timestamp(), enddate.timestamp()
    precision_dct = {'1H':3600, '24H':3600*24, 'hour': 3600, 'daily': 3600*24}
    runs, rest = divmod((enddate_unix-startdate_unix)/precision_dct[precision], 1000)
    rest = int(math.ceil(rest))
    url = "http://api-dev.augmento.ai/v0.1/events/aggregated"
    params = {
      "source" : source,
      "coin" : currency,
      "bin_size" : precision,
      "count_ptr" : 1000,
      "start_ptr" : 0,
      "start_datetime": convert_datetime(startdate),
      "end_datetime" : convert_datetime(enddate)} # "start_datetime" : "2019-05-01T00:00:00Z",
    output = pd.DataFrame()
    print(params)
    if runs != 0:
        for run in range(int(runs)):
            r = requests.request("GET", url, params=params)
            if r.status_code != 200:
                raise RuntimeError(r.json()['error']['message']) #resp.json()['error']['message']
            tmp = pd.DataFrame(r.json())
            tmp = pd.DataFrame(tmp['counts'].tolist(),index = tmp.datetime, columns = topics)
            output = pd.concat([output, tmp])
            params.update({'start_datetime': output.index.max()})
    #params['start_datetime'] = startdate
    params.update({'count_ptr': rest})
    print(params)
    r = requests.request("GET", url, params=params)
    if r.status_code != 200:
        raise RuntimeError(r.json()['error']['message'])
    elif len(r.json()) == 0:
        pass
    else:
        tmp = pd.DataFrame(r.json())
        tmp = pd.DataFrame(tmp['counts'].tolist(),index = tmp.datetime, columns = topics)
        output = pd.concat([output, tmp[1:]])
    # output = output.drop_duplicates()
    return output


def get_daily_data_augmento(currency='BTC', startdate=None, enddate=None, local_root_directory=None, target_present=False):
    currency_dict = {'BTC': 'bitcoin', 'ETH': 'ethereum'}
    frequence = 'daily'

    twitter = get_events_augmento(currency=currency_dict[currency], startdate=startdate, enddate=enddate, source='twitter', precision='24H', local_root_directory=local_root_directory)
    reddit = get_events_augmento(currency=currency_dict[currency], startdate=startdate, enddate=enddate, source='reddit', precision='24H', local_root_directory=local_root_directory)
    bitcointalk = get_events_augmento(currency=currency_dict[currency], startdate=startdate, enddate=enddate, source='bitcointalk', precision='24H', local_root_directory=local_root_directory)

    twitter = twitter.reset_index()
    twitter['datetime'] = pd.to_datetime(twitter.datetime)
    twitter = twitter.set_index(twitter.datetime)
    twitter.index = twitter.index.tz_convert(None)
    twitter = twitter.drop(columns=['datetime'])
    twitter = twitter.add_prefix('twitter_')

    reddit = reddit.reset_index()
    reddit['datetime'] = pd.to_datetime(reddit.datetime)
    reddit = reddit.set_index(reddit.datetime)
    reddit.index = reddit.index.tz_convert(None)
    reddit = reddit.drop(columns=['datetime'])
    reddit = reddit.add_prefix('reddit_')

    bitcointalk = bitcointalk.reset_index()
    bitcointalk['datetime'] = pd.to_datetime(bitcointalk.datetime)
    bitcointalk = bitcointalk.set_index(bitcointalk.datetime)
    bitcointalk.index = bitcointalk.index.tz_convert(None)
    bitcointalk = bitcointalk.drop(columns=['datetime'])
    bitcointalk = bitcointalk.add_prefix('bitcointalk_')

    aggregated_df = twitter.copy()
    aggregated_df = aggregated_df.merge(reddit, left_index=True, right_index=True)
    aggregated_df = aggregated_df.merge(bitcointalk, left_index=True, right_index=True)

    # start_datetime = datetime.datetime.fromtimestamp(convert_string(startdate))
    # end_datetime = datetime.datetime.fromtimestamp(convert_string(enddate))

    if target_present:
        freqly_pkl_file_name_suffix = '_' + frequence + '_returns.pkl'
        refetch_all = True
        starting_date = startdate
        running_date = enddate
        underlying_df = napoleon_connector.fetch_crypto_daily_data(ssj=currency, local_root_directory=local_root_directory,
                                                               daily_return_pkl_filename_suffix=freqly_pkl_file_name_suffix,
                                                               refetch_all=refetch_all,
                                                               daily_crypto_starting_day=starting_date,
                                                               daily_crypto_ending_day=running_date)
        underlying_df = underlying_df.drop(columns=['volumeto'])
        underlying_df = underlying_df.rename(columns={'volumefrom': 'volume'})
        aggregated_df = aggregated_df.merge(underlying_df, left_index=True, right_index=True)

    dates_stub = startdate.strftime('%d_%b_%Y') + '_' + enddate.strftime('%d_%b_%Y')
    augmento_suffix = '_augmento.pkl'
    pickle_saving_path = local_root_directory + currency + '_' + frequence + '_' + dates_stub + augmento_suffix
    aggregated_df.to_pickle(pickle_saving_path)
    return aggregated_df


def get_hourly_data_augmento(currency='BTC', startdate=None, enddate=None, local_root_directory=None, target_present=False):
    currency_dict = {'BTC': 'bitcoin', 'ETH': 'ethereum'}
    frequence = 'hourly'

    twitter = get_events_augmento(currency=currency_dict[currency], startdate=startdate, enddate=enddate, source='twitter', precision='1H', local_root_directory=local_root_directory)
    reddit = get_events_augmento(currency=currency_dict[currency], startdate=startdate, enddate=enddate, source='reddit', precision='1H', local_root_directory=local_root_directory)
    bitcointalk = get_events_augmento(currency=currency_dict[currency], startdate=startdate, enddate=enddate, source='bitcointalk', precision='1H', local_root_directory=local_root_directory)

    twitter = twitter.reset_index()
    twitter['datetime'] = pd.to_datetime(twitter.datetime)
    twitter = twitter.set_index(twitter.datetime)
    twitter.index = twitter.index.tz_convert(None)
    twitter = twitter.drop(columns=['datetime'])
    twitter = twitter.add_prefix('twitter_')

    reddit = reddit.reset_index()
    reddit['datetime'] = pd.to_datetime(reddit.datetime)
    reddit = reddit.set_index(reddit.datetime)
    reddit.index = reddit.index.tz_convert(None)
    reddit = reddit.drop(columns=['datetime'])
    reddit = reddit.add_prefix('reddit_')

    bitcointalk = bitcointalk.reset_index()
    bitcointalk['datetime'] = pd.to_datetime(bitcointalk.datetime)
    bitcointalk = bitcointalk.set_index(bitcointalk.datetime)
    bitcointalk.index = bitcointalk.index.tz_convert(None)
    bitcointalk = bitcointalk.drop(columns=['datetime'])
    bitcointalk = bitcointalk.add_prefix('bitcointalk_')

    aggregated_df = twitter.copy()
    aggregated_df = aggregated_df.merge(reddit, left_index=True, right_index=True)
    aggregated_df = aggregated_df.merge(bitcointalk, left_index=True, right_index=True)

    # start_datetime = datetime.datetime.fromtimestamp(convert_string(startdate))
    # end_datetime = datetime.datetime.fromtimestamp(convert_string(enddate))

    if target_present:
        freqly_pkl_file_name_suffix = '_' + frequence + '_returns.pkl'
        refetch_all = True
        starting_date = startdate
        running_date = enddate
        underlying_df = napoleon_connector.fetch_crypto_hourly_data(ssj = currency, local_root_directory=local_root_directory, hourly_return_pkl_filename_suffix = freqly_pkl_file_name_suffix, refetch_all = refetch_all,daily_crypto_starting_day=starting_date,daily_crypto_ending_day= running_date)
        underlying_df = underlying_df.drop(columns=['volumeto'])
        underlying_df = underlying_df.rename(columns={'volumefrom': 'volume'})
        aggregated_df = aggregated_df.merge(underlying_df, left_index=True, right_index=True)

    dates_stub = startdate.strftime('%d_%b_%Y') + '_' + enddate.strftime('%d_%b_%Y')
    augmento_suffix = '_augmento.pkl'
    pickle_saving_path = local_root_directory + currency + '_' + frequence + '_' + dates_stub + augmento_suffix
    aggregated_df.to_pickle(pickle_saving_path)
    return aggregated_df